import logging
from .client import JellyClient

LOG = logging.getLogger(__name__)
LOG.propagate = False

global __jelly__
__jelly__ = None
def init (*path, **config):
  global __jelly__
  route = []
  for p in path:
    if not isinstance(p, str):
      raise ValueError('path arguments must be string')
    for r in p.split('/'):
      if r:
        route.append(r)
  
  if not __jelly__:
    __jelly__ = JellyClient()
  
  scope = '.'.join(route[:3])
  __jelly__.session(scope)
  if config:
    __jelly__.configure(**config)
  return __jelly__

def serve ():
  global __jelly__
  __jelly__.serve()

def render (obj):
  global __jelly__
  __jelly__.render(obj)




import time
from collections.abc import Iterable

PRIMITIVE_TYPES = ['int', 'str', 'bool', 'float', 'long', 'NoneType']

def is_primitive (value):
  return type(value).__name__ in PRIMITIVE_TYPES

def is_iterable (value):
  return isinstance(value, Iterable)

def is_list (value):
  return isinstance(value, list) or isinstance(value, tuple)

def is_dict (value):
  return isinstance(value, dict)

def is_module (value):
  return type(value).__name__ == 'module'

def is_builtin (value):
  return is_primitive(value) or is_iterable(value) or is_dict(value) or is_module(value)

def is_instance (value):
  return not is_builtin(value)

def flatten (alist, result=None):
  if result is None:
    result = []
  for item in alist:
    if is_list(item):
      flatten(item, result)
    elif is_primitive(item):
      result.append(item)
    else:
      raise Exception('cannot flatten type: ' + type(item).__name__)
  return result

from uuid import uuid4
def now ():
  return int(time.time() * 1000)

def uid ():
  return str(uuid4())

def function_name (fn):
  klass = fn.__self__.__class__.__name__
  return "%s.%s" % (klass, fn.__name__)

def getpath (obj, path):
  if isinstance(path, str):
    path = path.split('.')
  result = obj
  for p in path:
    if is_instance(result):
      result = getattr(result, p)
    elif is_dict(result):
      result = result[p]
    elif is_iterable(result):
      result = result[int(p)]
    else:
      raise ValueError('%s does not exist' % p)
  return result

__ID__ = '__id__'
def object_key (obj):
  if hasattr(obj, __ID__):
    result = getattr(obj, __ID__)
    if callable(result):
      return str(result())
    return str(result)
  return str(id(obj))




import ssl, json, time, queue, atexit, logging, threading
import paho.mqtt.client as paho


LOG = logging.getLogger(__name__)


class MQTTJSONMessage:
  def __init__ (self, topic, payload):
    self.topic = topic
    self.ts = time.time() * 1000.0
    self.payload = json.loads(payload)

class MQTTSubscriber:
  def __init__ (self, mqtt, topic):
    self.mqtt = mqtt
    self.topic = topic
    self.queue = queue.Queue()

  def add_message (self, message):
    self.queue.put(message)
  
  def subscribe (self):
    LOG.debug('subscribe to: %s', self.topic)
    self.mqtt.subscribe([(self.topic + '/#', 1,)])

  def is_subscribed (self, topic):
    return topic.startswith(self.topic)
  
  def __iter__ (self):
    while not self.queue.empty():
      msg = self.queue.get()
      yield MQTTJSONMessage(msg.topic, msg.payload)

  def consume (self, sleep=0.01):
    while 1:
      for msg in self:
        yield msg
      time.sleep(sleep)


class MQTTMessage:
  def __init__ (self, topic, payload):
    self.topic = topic
    self.payload = payload
    self.publish_info = None
  
  def publish (self, client):
    self.publish_info = client.publish(self.topic, self.payload)



class MQTT:
  def __init__ (self, host, port, client_id, use_ssl=False, transport='tcp'):
    self.use_ssl = use_ssl
    self.client_id = client_id
    self.mqtt_host = host
    self.mqtt_port = port
    self.subscriptions = []
    self.is_connected = None
    self.transport = transport

    self.client = paho.Client(transport=self.transport, client_id=self.client_id)
    self.client.on_message = self.on_message
    self.client.on_connect = self.on_connect
    self.client.on_disconnect = self.on_disconnect
    self.client.max_inflight_messages_set(0)

    self.out_messages = queue.Queue()

    if self.use_ssl:
        ssl_ctx = ssl.create_default_context()
        ssl_ctx.check_hostname = False
        ssl_ctx.verify_mode = ssl.CERT_NONE
        self.client.tls_set_context(ssl_ctx)
        #self.client.tls_insecure_set(True)

    atexit.register(self.on_exit)

    self.thread = threading.Thread(target=self.main)
    self.thread.daemon = True
    self.thread.start()

  def _ensure_connection (self):
    while 1:
      try:
        result = self.client.connect(self.mqtt_host, self.mqtt_port)
        self.client.loop_start()
      except:
        LOG.warning('failed to connect to server')
        time.sleep(5)
      else:
        break

  def main (self):
    self._ensure_connection()
    while 1:
      time.sleep(0.02)
      if not self.is_connected or self.out_messages.empty():
        continue
      msg = self.out_messages.get()
      msg.publish(self.client)
            
  def subscribe (self, topic):
    sub = MQTTSubscriber(self.client, topic)
    self.subscriptions.append(sub)
    return sub

  def on_connect (self, *args, **kw):
    LOG.debug('connected: %s', self.client_id)
    self.is_connected = True
    for sub in self.subscriptions:
      sub.subscribe()
    
  def on_disconnect (self, *args, **kw):
    LOG.debug('disconnected')
    self.is_connected = False
    
  def on_message (self, client, _, msg):
    LOG.debug('received: %s', msg.topic)
    for sub in self.subscriptions:
      if not sub.is_subscribed(msg.topic):
        continue
      sub.add_message(msg)
    
  def publish (self, topic, msg):
    self.out_messages.put(
      MQTTMessage(topic, json.dumps(msg))
    )
  
  def on_exit (self):
    # TODO: save unsent messages to file until confirmed
    while not self.out_messages.empty():
      LOG.info('waiting for pending messages before exit..')
      time.sleep(1) 


import time, sys, logging, os

from .transport import MQTT
from .logger import JellyLogger, LoggingLevel
from .util import uid
from .session import JellySession

LOG = logging.getLogger(__name__)

class SessionHandler:
  def __init__ (self):
    self.process = None
    self.queue = None
    
class JellyClient:
  def __init__ (self, version='v1', host='ws.web-jelly.com', port=1883):
    self.version = version
    self.client_id = uid()    
    self.config = dict(
      host=host, port=port
    )

    self.transport = None    
    self.sessions = {}
    
  def session (self, scope):
    if not scope in self.sessions:
      self.sessions[scope] = JellySession(self, scope)
    return self.sessions[scope]
  
  def serve (self):
    while 1:
      time.sleep(0.1)
    
  def kill (self):
    self.transport.wait_for_publish()
    os._exit(0)
  
  def subscribe_broadcast (self):
    return self.transport.subscribe("%s/%s" % (self.version, 'broadcast'))

  def configure (self, **kw):
    self.config.update(**kw)
    self.transport = MQTT(self.config['host'], self.config['port'], self.client_id)
    app_path = os.path.abspath(self.config.get('app', '.'))
    sys.path.append(app_path)
    for id, session in self.sessions.items():
      session.configure(**kw)
    
  def publish (self, topic, msg):
    self.transport.publish(topic, msg)

  def subscribe (self, topic):
    return self.transport.subscribe(topic)





import traceback, json, types
from .util import (
  is_primitive, is_module, is_dict, is_iterable, is_instance, is_list, is_builtin
)





class ComponentElement:
  def __init__ (self, engine):
    self.engine = engine
    self.handler = None
    self._elements = {}
    self.subcomponents = {}
    self._active_elements = set()

  def diff_elements (self, e1, e2):
    if e1.type != e2.type:
      return True
    if type(e1) is not type(e2):
      return False
    if isinstance(e1, TextElement):
      return e1.value != e2.value
    elif isinstance(e1, Element):
      keys = e1.props.keys() & e2.props.keys()
      for k in keys:
        if e1.props.get(k) != e2.props.get(k):
          return True
    return False

  def create_element (self, parent, index, element):
    #key = parent + '.' + str(index)
    item = {'id': element.key, 'index': index, 'parent': parent.key, 'element': element.as_dict()}
    self._elements[item['id']] = element
    return item

  def open_element (self, element):
    return {'action': 'OPEN', 'element': element.as_dict()}

  def close_element (self, element):
    return {'action': 'CLOSE', 'element': element.as_dict()}
  
  def render_to_string (self, template):
    result = ''
    listeners = {}
    for item in self.render('', template):
      e = item['element']
      if e['type'] == 'TEXT':
        result += str(e['value'])
        continue
      props = []

      if item['action'] == 'CLOSE':
        result += '</%s>' % e['type']
        continue
      props.append('data-key="%s"' % e['key'])
      for k, v in e['props'].items():
        props.append('%s="%s"' % (k, v))
      
      listeners.setdefault(e['key'], {})
      listeners[e['key']].update(e['listeners'])
        
      for k, v in e['listeners'].items():
        props.append('''on%s="ON_EVENT('%s', this);"''' % (k, k))
      result += '<%s%s%s>' % (
        e['type'], ' ' if len(props) else '', ' '.join(props)
      )
    result += '''
<script>
register_listeners(%s)
</script>
''' % listeners
    return result

  def render (self, root, template):
    #self.handler = handler
    self._elements = {}
    root_element = Element(self, None, 'div', [])
    root_element.key = root
    result = template
    try:
      if isinstance(result, Element):
        yield from self.render_element(root_element, 0, result)
      elif is_primitive(result):
        yield from self.render_element(root_element, 0, result)
      elif isinstance(result, (types.GeneratorType, list, tuple,)):
        for index, element in enumerate(result):
          yield from self.render_element(root_element, index, element)
      else:
        raise Exception('unknown render result: ' + repr(result))
      yield from self.render_subcomponents()
    except Exception as err:
      yield from self.render_element(root_element, 0, str(err))

  def render_element (self, parent, index, obj):
    element = None
    if is_primitive(obj):
      element = TextElement(parent, obj)
    elif isinstance(obj, Node):
      element = obj
    else:
      element = ComponentRef(parent, obj)
    element.key = "%s.%s" % (parent.key, index)
    #yield self.create_element(parent, index, element)
    yield self.open_element(element)
    if hasattr(element, 'children'):
      for child_index, child in enumerate(self.list_children(element, element.children)):
        #child.parent = element
        #parent_key = "%s.%s" % (parent, index)
        yield from self.render_element(element, child_index, child)
      yield self.close_element(element)

  def list_children (self, element, children):
    for item in children:
      if is_list(item):
        yield from self.list_children(element, item)
      #elif is_primitive(item):
      #  element = TextElement(element, item)
      else:
        yield item
      #elif is_primitive(item):
      #  yield TextElement(element, item)
      #elif isinstance(item, Node):
      #  yield item
      #else:
      #  yield ComponentRef(element, item)

  def render_subcomponents (self):
    comps = {k: c for (k, c) in self._elements.items() if c.type == 'REF'}
    for key, comp in comps.items(): 
      if not key in self.subcomponents:
        self.subcomponents[key] = ComponentElement(self.engine)
      subcomp = self.subcomponents[key]
      yield from subcomp.render(key, comp.component)


  # API METHODS
  @property
  def elements (self):
    return Elements(self)

  def emit (self, type, **kw):
    return Event(type, **kw)


class Event:
  def __init__ (self, id, type, **kw):
    self.id = id
    self.type = type
    self.payload = kw

  def trigger (self, payload):
    self._on_trigger(payload)

  def on_trigger (self, callback):
    self._on_trigger = callback
    return self
  
  def emit (self, **values):
    return EventTrigger(self, values)


class EventTrigger:
  def __init__ (self, event, values):
    self.event = event
    self.payload = values
  
  def as_dict (self):
    return {'id': self.event.id, 'type': self.event.type, 'payload': self.payload}


class Elements:
  def __init__ (self, ctx):
    self.ctx = ctx
  def __getattr__ (self, name):
    return ElementSignature(self.ctx, None, name)

class ElementSignature:
  def __init__ (self, ctx, parent, type, **props):
    self.ctx = ctx
    self.type = type
    self.parent = parent
    
  def __call__ (self, id=None, **props):
    return Element(self.ctx, self.parent, self.type, id, **props)

class Node:
  def __init__ (self, type, parent):
    self.key = ''
    self.type = type
    self.parent = parent

  @property
  def id (self):
    return ''
  
  @property
  def test (self):
    p = self.parent
    keys = [self.id]
    while p:
      keys.insert(0, p.id)
      p = p.parent
    return '/'.join(keys)

class TextElement(Node):
  def __init__ (self, parent, value):
    super().__init__('TEXT', parent)
    self.value = value

  @property
  def id (self):
    return str(hash(self.value))
  
  def as_dict (self):
    return {'type': self.type, 'value': self.value}

class ComponentRef(Node):
  def __init__ (self, parent, component):
    super().__init__('REF', parent)
    self.component = component

  @property
  def id (self):
    return self.component.__class__.__name__

  def as_dict (self):
    return {'type': self.type}
  
def to_dict (x):
  if is_primitive(x) or isinstance(x, dict):
    return x
  if hasattr(x, 'as_dict'):
    return x.as_dict()
  return None  

class Element(Node):
  def __init__ (self, ctx, parent, type, id, **props):
    super().__init__(type, parent)
    self.ctx = ctx
    self.props = props
    self.children = [] # ctx.list_children(self, children)
    self.listeners = {}
    if id:
      self.props['id'] = id.strip('#')

  @property
  def id (self):
    return self.props.get('id', '')

  def __serialize__ (self, serializer):
    props = {k: parse_prop(p) for (k, p) in self.props.items()}
    listeners = {k: parse_trigger(v) for (k, v) in self.listeners.items()}
    return dict(type='html', value=dict(
      type=self.type, props=props, 
      listeners=listeners, key=self.key,
      children=[serializer(c) for c in self.children]
    ))

  def __str__ (self):
    return "<%s>%s</%s>" % (self.type, 
      '\n'.join([str(c) for c in self.children]), self.type
    )

  def on (self, **listeners):
    #for k, v in listeners.items():
    #  if isinstance(v, types.FunctionType):
    #    v = Event()
    self.listeners.update(listeners)
    return self

  def click (self, trigger):
    return self.on(click=trigger)

  def style (self, **props):
    # TODO: do not overwrite any existing styles, unless it's a re-render
    if not 'style' in self.props:
      self.props['style'] = {}
    self.props['style'].update(**props)
    return self
    style_string = ''
    for k, v in props.items():
      key = k.replace('_', '-')
      style_string += "%s:%s;" % (key, v)
    self.props['style'] = style_string
    return self

  def classes (self, *classnames, **classes):
    names = {k for (k,v) in classes.items() if v} | set(classnames)
    new_names = set(self.props.get('class', '').split(' ')) | names
    #self.props['class'] = ' '.join(new_names)
    self.props['className'] = ' '.join(new_names)
    return self

  def __call__ (self, *children):
    self.children = self.ctx.list_children(self, children)
    return self


# TODO: this should be element-based
#  for example, "line" element will expect "point" elements in "points" attribute
#  have some "extract_elements("points")" maybe
def parse_prop (prop):
  if hasattr(prop, 'as_dict'):
    return prop.as_dict()['props']
  elif isinstance(prop, list):
    return [parse_prop(p) for p in prop]
  return prop

def parse_trigger (trigger):
  return trigger.as_dict()



import uuid

class Template:
  def __init__ (self, obj, element):
    self.obj = obj
    self.element = element

  def __serialize__ (self, serialize):
    return {
      'type': 'template', 'class': self.obj.__class__.__name__,
      'id': str(id(self.obj)), 'value': serialize(self.element)
    }

class RenderContext:
  def __init__ (self):
    #self.id = id
    #self.mqtt = mqtt
    self.component = ComponentElement(None)
    self.events = {}
    self.state = {}

  def set_state (self, **values):
    self.state.update(values)

  # API METHODS
  @property
  def elements (self):
    return Elements(self.component)

  def trigger (self, evt):
    self.events[evt['id']].trigger(evt['payload'])

  def event (self, type, **kw):
    id = str(uuid.uuid4())
    evt = Event(id, type, **kw)
    self.events[id] = evt
    return evt

  def render (self, template):
    html = self.component.render_to_string(template)
    return html
    #self.mqtt.publish('v1/output/' + self.id, dict(
    #  type='HTML', event='RENDER', message=html, request_id=self.id
    #))


import multiprocessing, logging, platform, threading, time
from .logger import JellyLogger, LoggingLevel

LOG = logging.getLogger(__name__)
ENV = multiprocessing.get_context("spawn")

def evaluate_script (session):
  _globals = {}
  _globals.update(**globals())
  _globals.update(
    info=logging.info, debug=logging.debug, error=logging.error
  )
  scope = {}
  code = session.get('code', '')
  try:
    exec(code, _globals, scope)
    if '__main__' in scope:
      scope['__main__']()
  except Exception as err:
    #LOG.error("%s", err)
    LOG.exception(err)
  finally:
    pass
    #LOG.log(LoggingLevel.STATUS, 'evaluate complete')

def linux_distribution():
  try:
    return platform.linux_distribution()
  except:
    return "N/A"

def os_information ():
  return dict(
    python_version=platform.python_version(),
    libc_version='-'.join(platform.libc_ver()),
    node=platform.node(),
    linux_distribution=linux_distribution(),
    system=platform.system(),
    release=platform.release(),
    machine=platform.machine(),
    platform=platform.platform(),
    version=platform.version(),
  )

class JellySession:
  def __init__ (self, client, scope):
    self.client = client
    self.scope = scope.lower()
    self.model = {}
    self.config = {}
    self.configured = False
    self.logging = JellyLogger(self)
    self.rpc = JellyRPC(self)
    
  #def inspect (self, object=None, paths=None):
  #  data = self.logging.inspect(object=object, paths=paths)
  #  if data:
  #    self.publish('inspect', data)
  
  
  def publish (self, method, payload, *args):
    topic = self.make_topic(method, args=list(args))
    payload.update(scope=self.scope)
    LOG.debug("%s -> %s", self.scope, topic)
    self.client.publish(topic, payload)

  def configure (self, **config):
    self.config = config
    self.logging.activate(**config.get('logging', {}))
    self.rpc.activate(**config.get('rpc', {}))
    self.introduce()
    LOG.info('process started: %s', self.scope)
    self.configured = True

  def subscribe_method (self, method):
    topic = self.make_topic(method, 'inbox')
    return self.client.subscribe(topic)
  
  def introduce (self):
    self.publish('status', dict(
      system=os_information(),
      metadata=self.config.get('metadata', {})
    ))

  def make_topic (self, method, group='outbox', args=None):
    user_type = 'client'
    rest = '' if not args else ('/' + '/'.join(map(str, args)))
    topic = "%s/%s/%s/%s/%s/%s%s" % (
      self.client.version, group, user_type, self.client.client_id, 
      self.scope, method, rest
    )
    return topic.lower()

  def set_model (self, model):
    if self.model == model:
      LOG.debug('no changes in session, skipping: %s', model['id'])
      return
    LOG.debug('session model updated')
    self.model = model
    if not self.configured:
      self.configure(**self.config)
    evaluate_script(model)
    #if self.process:
    #  self.process.kill()
    #self.queue = ENV.Queue()
    #self.process = ENV.Process(target=evaluate_script, args=(self.queue, session))
    #self.process.start()
    
  def evaluate (self, code):
    pass
    #LOG.log(LoggingLevel.STATUS, 'starting evaluate')
    
  def register_session (self, model):
    session = self.client.session(model['scope'])
    if session is self:
      return
    session.config = self.config.copy()
    session.set_model(model)
    
  def subscribe_broadcast (self):
    return self.client.subscribe_broadcast()


# == RPC ==
class Action:
  def __init__ (self, client, action_id, handler):
    self.client = client
    self.id = action_id
    self.handler = handler
  
  def execute (self, param):
    self.client.execute_action(self.id, self.handler, param)

class JellyRPC:
  def __init__ (self, session):
    self.session = session
    self.active_calls = {}
    self.commands = {}
    self._main_thread = None
    
  def set_command (self, name, args, handler):
    self.commands[name] = dict(
      args=args, handler=handler
    )

  def activate (self, **kwargs):
    self._main_thread = threading.Thread(target=self._main_loop)
    self._main_thread.daemon = True
    self._main_thread.start()

  def _main_loop (self):
    requests = self.session.subscribe_method('rpc')
    broadcasts = self.session.subscribe_broadcast()
    while 1:
      for breq in broadcasts:
        self.session.introduce()
      for req in requests:
        try:
          self._handle_request(req)
        except Exception as err:
          LOG.exception(err)
      time.sleep(0.01)

  def _handle_request (self, req):
    LOG.info('handle request: %s, %s', self.session.scope, req.payload)
    type = req.payload.get('type', '').lower().strip()
    args = req.payload.get('payload', {})
    cmd = self.commands.get(type)
    if type == '__session__':
      self.session.register_session(args)      
    #elif type == '__inspect__':
    #  self.session.inspect(**args)
    elif not cmd:
      raise RuntimeError('no such command: %s' % type)
    else:
      cmd['handler'](**args)
    


import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="web-jelly",
    version="0.1.1",
    author="Jellyfish",
    author_email="hello@web-jelly.com",
    description="web-jelly.com python client",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://web-jelly.com",
    project_urls={
        #'Documentation': 'https://packaging.python.org/tutorials/distributing-packages/',
        "Bug Tracker": "https://gitlab.com/hello590/jelly-python-client/-/issues",
        #'Funding': 'https://donate.pypi.org',
        #'Say Thanks!': 'http://saythanks.io/to/example',
        'Source': 'https://gitlab.com/hello590/jelly-python-client'
        #'Tracker': 'https://github.com/pypa/sampleproject/issues',
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    install_requires=[
        'paho-mqtt'
    ],
    python_requires=">=3.6",
)


import inspect, time, sys, builtins, logging, traceback, re, hashlib, weakref
from .util import (
  is_iterable, is_primitive, is_dict, is_instance, is_iterable,
  is_module, is_builtin, uid, now, function_name, getpath, object_key
)

LOG = logging.getLogger(__name__)

TYPE_KEY = '__type__'
EXIST_KEY = '__exist__'

class TYPE:
  LOG='LOG'
  EVENT='EVENT'
  OBJECT='OBJECT'

class LoggingLevel:
  STATUS = 60
  CRITICAL = 50
  ERROR = 40
  WARNING = 30
  INFO = 20
  DEBUG = 10
  NOTSET = 0

def log_level_name (level):
  if level == LoggingLevel.STATUS:
    return 'STATUS'
  elif level == LoggingLevel.CRITICAL:
    return 'CRITICAL'
  elif level == LoggingLevel.ERROR:
    return 'ERROR'
  elif level == LoggingLevel.WARNING:
    return 'WARNING'
  elif level == LoggingLevel.INFO:
    return 'INFO'
  elif level == LoggingLevel.NOTSET:
    return 'NOTSET'
  return ''

def get_level (name):
  for k, v in LoggingLevel.__dict__.items():
    if k == name:
      return v
  return LoggingLevel.NOTSET

class LogType:
  LOG=0
  EVENT=1

class JellyfishLoggingHandler(logging.NullHandler):
  def __init__ (self, client):
    super(JellyfishLoggingHandler, self).__init__()
    self.client = client
    
  def emit(self, record):
    publish_logging_record(self.client, record)


def create_stack (stack):
  info = inspect.getinnerframes(stack)
  stack_data = []
  for frame in info:
    frame_data = {
      'source': '' if not frame.code_context else ''.join(list(frame.code_context)),
      'line': frame.lineno, 'file': frame.filename, 'function': frame.function
    }
    stack_data.append(frame_data)
  return stack_data

def publish_logging_record (client, record):
  if record.name.startswith('jelly'):
    return
  
  record.levelname = log_level_name(record.levelno)
  message = record.getMessage()
  try:
    stack = None
    log_type = LogType.LOG
    log_hash = None
    msg = record.msg
    args = record.args
    
    if record.exc_info:
      exc = traceback.format_exception(*record.exc_info)
      exception = record.exc_info[1]
      exc_class = record.exc_info[0]
      msg = "%s: %s" % (exc_class.__name__, str(exception)) # ''.join(exc)
      log_type = LogType.EVENT
      stack = create_stack(record.exc_info[2])
      hash_value = exc_class.__name__ + ':' + '-'.join([s['source'] for s in stack])
      
      hash_object = hashlib.md5(hash_value.encode())
      log_hash = str(hash_object.hexdigest())
      args = [{
        TYPE_KEY: TYPE.EVENT, 'id': log_hash, 'message': msg, 'details': {}
      }]
      
    elif not isinstance(msg, str):
      args = [msg]
      msg = "%s"
    
    if args and not isinstance(args, (tuple, list,)):
      args = [args]  
    items = [client.serializer.serialize(arg) for arg in args]
    
    shape = "%s:%s:%s:%s:%s" % (
      len(message), len(list(re.finditer("\r\n?|\n", message))),
      len(msg), len(list(re.finditer("\r\n?|\n", message))),
      len(args)
    )
    client.publish_log(dict(__type__=TYPE.LOG,
      task=record.extra.get('task'),
      name=record.name, msg=msg, level=record.levelname,
      levelno=record.levelno, ts=time.time()*1000.0,
      thread=record.thread, threadName=record.threadName,
      process=record.process, module=record.module,
      lineno=record.lineno, pathname=record.pathname,
      args=items, message=message, shape=shape,
      stack=stack, type=log_type, hash=log_hash
    ))
  except Exception as err:
    LOG.error('failed to parse log: %s', err)

class TaskLogger:
  def __init__ (self, logger, id=None, total=0, completed=0):
    self.logger = logger
    self.id = id or uid()
    self.total = total
    self.completed = completed

  def log (self, level, msg, *args, **kwargs):
    kwargs.setdefault('extra', {})
    if 'completed' in kwargs:
      self.completed = kwargs['completed']
    kwargs['extra']['task'] = dict(
      id=self.id, completed=self.completed, total=self.total
    )
    self.logger.log(level, msg, *args, **kwargs)

  def info (self, msg, *args, **kwargs):
    self.log(LoggingLevel.INFO, msg, *args, **kwargs)
  def error (self, msg, *args, **kwargs):
    self.log(LoggingLevel.ERROR, msg, *args, **kwargs)
  def debug (self, msg, *args, **kwargs):
    self.log(LoggingLevel.DEBUG, msg, *args, **kwargs)
  def warn (self, msg, *args, **kwargs):
    self.log(LoggingLevel.WARNING, msg, *args, **kwargs)


def object_ref (obj):
  return {'id': str(id(obj)), 'class': type(obj).__name__}

class JellyLogger:

  def __init__ (self, client):
    self.client = client
    self.active_request = None
    self.level = LoggingLevel.NOTSET
    self.current_print_line = None
    self.serializer = ObjectSerializer()

  def task (self, logger, id=None, completed=None, total=None):
    return TaskLogger(logger, id=id, completed=completed, total=total)


  def inspect (self, object=None, paths=None):
    if not object:
      objects = self.serializer.list_objects()
      return {
        EXIST_KEY: True, 
        'objects': [object_ref(obj) for obj in objects]
      }
    
    inst = self.serializer.get_object(object)
    if not inst:
      return {EXIST_KEY: False, 'object': object}

    if not paths:
      return {EXIST_KEY: True, 'object': object, 'value': self.serializer.serialize(inst)}

    result = {
      EXIST_KEY: True, 'object': object, 'values': {}
    }
    for path in paths:
      try:
        value = getpath(inst, path)
        result['values'][path] = {
          EXIST_KEY: True, 'value': self.serializer.serialize(value)
        }
      except:
        result['values'][path] = {EXIST_KEY: False}
    return result

  def activate (self, level=LoggingLevel.DEBUG):
    # NOTE:
    #   for some reason, adding a handler messes with basicConfig somehow
    #   have to use logging factory to intercept logs 
    self.level = get_level(level)
    old_factory = logging.getLogRecordFactory()
    def record_factory(*args, **kwargs):
      record = old_factory(*args, **kwargs)
      if not hasattr(record, 'extra'):
        record.extra = {}
      record.extra.update(**kwargs.get('extra', {}))
      publish_logging_record(self, record)
      return record
    logging.setLogRecordFactory(record_factory)
    
    print_logger = logging.getLogger('print')
    def new_print (*args, **kwargs):
      end = kwargs.get('end')
      log_msg = " ".join(["%s" for _ in range(len(args))])
      print_logger.info(log_msg, *args)
    
    old_excepthook = sys.excepthook
    def excepthook (exc_type, exc_value, exc_traceback):
      if exc_type is not KeyboardInterrupt:
        logging.error('', exc_info=(exc_type, exc_value, exc_traceback))
      old_excepthook(exc_type, exc_value, exc_traceback)
    sys.excepthook = excepthook
    #builtins.print = new_print    
    
    # NOTE: do not do this: (test with basicConfig)
    # logging.root.addHandler(JellyfishLoggingHandler(self))
    
  def publish_log (self, log):
    level = log.get('levelno')
    log_level = self.level
    if not isinstance(level, int) or not isinstance(log_level, int):
      return
    if log_level <= level:
      log['_id'] = str(uid())
      try:
        self.client.publish('log', log, log['type'], log['levelno'])
      except Exception as err:
        LOG.error('failed to publish log: %s', err)

  def set_active_request (self, req_id):
    self.active_request = req_id

  

class ObjectSerializer:
  def __init__ (self):
    self._objects = {}

  def _register (self, obj):
    try:
      obj_key = object_key(obj)
      if obj_key not in self._objects:
        self._objects[obj_key] = weakref.ref(obj, self._on_object_deleted)
    except Exception as err:
      LOG.error('failed to register object: %s', obj)

  def _on_object_deleted (self, obj):
    pass

  def get_object (self, obj_id):
    obj = self._objects[obj_id]
    if obj:
      return obj.ref()
    return obj

  def serialize (self, value):
    result = None
    if is_primitive(value):
      result = value
    elif is_dict(value):
      result = {k: self.serialize(v) for k, v in value.items()}
    elif is_iterable(value):
      result = [self.serialize(v) for v in value]
    elif hasattr(value, '__serialize__'):
      result = value.__serialize__(self.serialize)
    elif is_instance(value) and hasattr(value, '__event__'):
      result = {
        TYPE_KEY: TYPE.EVENT
      }
      result.update(**value.__event__())      
    #elif isinstance(value, Exception):
    #  result = {
    #    TYPE_KEY: 'event'
    #  }
    #  message = type(value).__name__ + ': ' + str(value)
    #  result.update(**dict(
    #    message=message,
    #    id=message, level=LoggingLevel.ERROR
    #  ))
    elif is_instance(value):
      if not isinstance(value, Exception):
        self._register(value)
      state = {}
      result = {TYPE_KEY: TYPE.OBJECT}
      if hasattr(value, '__state__'):
        state = self.serialize(value.__state__())
      result.update(**{
        'class': type(value).__name__,
        'id': object_key(value),
        'message': str(value),
        'state': state
      })
    else:
      result = {TYPE_KEY: 'unknown'}
    return result
